/*********************************************************************
** ATA LOW LEVEL I/O DRIVER -- ATAIOPIO.C
**
** by Hale Landis (landis@sugs.tware.com)
**
** There is no copyright and there are no restrictions on the use
** of this ATA Low Level I/O Driver code.  It is distributed to
** help other programmers understand how the ATA device interface
** works and it is distributed without any warranty.  Use this
** code at your own risk.
**
** This code is based on the ATA-2, ATA-3 and ATA-4 standards and
** on interviews with various AT controller and drive designers.
**
** This code has been run on many AT (IDE) drives and
** MFM/RLL controllers.  This code may be a little
** more picky about the status it sees at various times.  A real
** BIOS probably would not check the status as carefully.
**
** Compile with one of the Borland C or C++ compilers.
** This module contains inline assembler code so you'll
** also need Borland's Assembler.
**
** This C source contains the low level I/O port IN/OUT functions.
*********************************************************************/

#include <dos.h>

#include "ataio.h"

#include "ataiopd.h"

/**************************************************************
**
** Host adapter base addresses.
**
**************************************************************/

unsigned int pio_base_addr1 = 0x1f0;
unsigned int pio_base_addr2 = 0x3f0;
unsigned int pio_memory_seg = 0;

unsigned int pio_reg_addrs[10];

unsigned char pio_last_write[10];
unsigned char pio_last_read[10];

/**************************************************************
**
** Set the host adapter i/o base addresses.
**
**************************************************************/

void pio_set_iobase_addr( unsigned int base1, unsigned int base2 )

{

   pio_base_addr1 = base1;
   pio_base_addr2 = base2;
   pio_memory_seg = 0;
   pio_reg_addrs[ CB_DATA ] = pio_base_addr1 + 0;  /* 0 */
   pio_reg_addrs[ CB_FR   ] = pio_base_addr1 + 1;  /* 1 */
   pio_reg_addrs[ CB_SC   ] = pio_base_addr1 + 2;  /* 2 */
   pio_reg_addrs[ CB_SN   ] = pio_base_addr1 + 3;  /* 3 */
   pio_reg_addrs[ CB_CL   ] = pio_base_addr1 + 4;  /* 4 */
   pio_reg_addrs[ CB_CH   ] = pio_base_addr1 + 5;  /* 5 */
   pio_reg_addrs[ CB_DH   ] = pio_base_addr1 + 6;  /* 6 */
   pio_reg_addrs[ CB_CMD  ] = pio_base_addr1 + 7;  /* 7 */
   pio_reg_addrs[ CB_DC   ] = pio_base_addr2 + 6;  /* 8 */
   pio_reg_addrs[ CB_DA   ] = pio_base_addr2 + 7;  /* 9 */
}

/**************************************************************
**
** Set the host adapter memory base addresses.
**
**************************************************************/

void pio_set_memory_addr( unsigned int seg )

{

   pio_base_addr1 = 0;
   pio_base_addr2 = 8;
   pio_memory_seg = seg;
   pio_reg_addrs[ CB_DATA ] = pio_base_addr1 + 0;  /* 0 */
   pio_reg_addrs[ CB_FR   ] = pio_base_addr1 + 1;  /* 1 */
   pio_reg_addrs[ CB_SC   ] = pio_base_addr1 + 2;  /* 2 */
   pio_reg_addrs[ CB_SN   ] = pio_base_addr1 + 3;  /* 3 */
   pio_reg_addrs[ CB_CL   ] = pio_base_addr1 + 4;  /* 4 */
   pio_reg_addrs[ CB_CH   ] = pio_base_addr1 + 5;  /* 5 */
   pio_reg_addrs[ CB_DH   ] = pio_base_addr1 + 6;  /* 6 */
   pio_reg_addrs[ CB_CMD  ] = pio_base_addr1 + 7;  /* 7 */
   pio_reg_addrs[ CB_DC   ] = pio_base_addr2 + 6;  /* 8 */
   pio_reg_addrs[ CB_DA   ] = pio_base_addr2 + 7;  /* 9 */
}

/*********************************************************/

unsigned char pio_inbyte( unsigned int addr )

{
   unsigned int regAddr;
   unsigned char uc;
   unsigned char far * ucp;

   regAddr = pio_reg_addrs[ addr ];
   if ( pio_memory_seg )
   {
      ucp = (unsigned char far *) MK_FP( pio_memory_seg, regAddr );
      uc = * ucp;
   }
   else
   {
      uc = (unsigned char) inportb( regAddr );
   }
   pio_last_read[ addr ] = uc;
   trc_llt( addr, uc, TRC_LLT_INB );

   /* ignore the Index bit in the Status register */
   if ( regAddr == CB_STAT || regAddr == CB_ASTAT )
      uc = uc & ( ~ CB_STAT_IDX );

   return uc;
}

/*********************************************************/

void pio_outbyte( unsigned int addr, unsigned char data )

{
   unsigned int regAddr;
   unsigned char far * ucp;

   regAddr = pio_reg_addrs[ addr ];
   if ( pio_memory_seg )
   {
      ucp = (unsigned char far *) MK_FP( pio_memory_seg, regAddr );
      * ucp = data;
   }
   else
   {
      outportb( regAddr, data );
   }
   pio_last_write[ addr ] = data;
   trc_llt( addr, data, TRC_LLT_OUTB );
}

/*********************************************************/

unsigned int pio_inword( unsigned int addr )

{
   unsigned int regAddr;
   unsigned int ui;
   unsigned int far * uip;

   regAddr = pio_reg_addrs[ addr ];
   if ( pio_memory_seg )
   {
      uip = (unsigned int far *) MK_FP( pio_memory_seg, regAddr );
      ui = * uip;
   }
   else
   {
      ui = inport( regAddr );
   }
   trc_llt( addr, 0, TRC_LLT_INW );
   return ui;
}

/*********************************************************/

void pio_outword( unsigned int addr, unsigned int data )

{
   unsigned int regAddr;
   unsigned int far * uip;

   regAddr = pio_reg_addrs[ addr ];
   if ( pio_memory_seg )
   {
      uip = (unsigned int far *) MK_FP( pio_memory_seg, regAddr );
      * uip = data;
   }
   else
   {
      outport( regAddr, data );
   }
   trc_llt( addr, 0, TRC_LLT_OUTW );
}

/*********************************************************/

void pio_rep_inword( unsigned int addrDataReg,
                     unsigned int bufSeg, unsigned int bufOff,
                     unsigned int wordCnt )

{
   unsigned int dataRegAddr;
   unsigned int far * uip1;
   unsigned int far * uip2;

   dataRegAddr = pio_reg_addrs[ addrDataReg ];
   if ( pio_memory_seg )
   {
      uip1 = (unsigned int far *) MK_FP( pio_memory_seg, dataRegAddr );
      uip2 = (unsigned int far *) MK_FP( bufSeg, bufOff );
      for ( ; wordCnt > 0; wordCnt -- )
      {
         * uip2 = * uip1;
         uip2 ++ ;
      }
   }
   else
   {
      asm   .286

      asm   push  ax
      asm   push  cx
      asm   push  dx
      asm   push  di
      asm   push  es

      asm   mov   ax,bufSeg
      asm   mov   es,ax
      asm   mov   di,bufOff

      asm   mov   cx,wordCnt
      asm   mov   dx,dataRegAddr

      asm   cld

      asm   rep   insw

      asm   pop   es
      asm   pop   di
      asm   pop   dx
      asm   pop   cx
      asm   pop   ax
   }
   trc_llt( addrDataReg, 0, TRC_LLT_INSW );
}

/*********************************************************/

void pio_rep_outword( unsigned int addrDataReg,
                      unsigned int bufSeg, unsigned int bufOff,
                      unsigned int wordCnt )

{
   unsigned int dataRegAddr;
   unsigned int far * uip1;
   unsigned int far * uip2;

   dataRegAddr = pio_reg_addrs[ addrDataReg ];
   if ( pio_memory_seg )
   {
      uip1 = (unsigned int far *) MK_FP( bufSeg, bufOff );
      uip2 = (unsigned int far *) MK_FP( pio_memory_seg, dataRegAddr );
      for ( ; wordCnt > 0; wordCnt -- )
      {
         * uip2 = * uip1;
         uip1 ++ ;
      }
   }
   else
   {
      asm   .286

      asm   push  ax
      asm   push  cx
      asm   push  dx
      asm   push  si
      asm   push  ds

      asm   mov   ax,bufSeg
      asm   mov   ds,ax
      asm   mov   si,bufOff

      asm   mov   cx,wordCnt
      asm   mov   dx,dataRegAddr

      asm   cld

      asm   rep   outsw

      asm   pop   ds
      asm   pop   si
      asm   pop   dx
      asm   pop   cx
      asm   pop   ax
   }
   trc_llt( addrDataReg, 0, TRC_LLT_OUTSW );
}

/* end ataiopio.c */
