/*********************************************************************
** ATA LOW LEVEL I/O DRIVER -- ATAIOREG.C
**
** by Hale Landis (landis@sugs.tware.com)
**
** There is no copyright and there are no restrictions on the use
** of this ATA Low Level I/O Driver code.  It is distributed to
** help other programmers understand how the ATA device interface
** works and it is distributed without any warranty.  Use this
** code at your own risk.
**
** This code is based on the ATA-2, ATA-3 and ATA-4 standards and
** on interviews with various AT controller and drive designers.
**
** This code has been run on many AT (IDE) drives and
** MFM/RLL controllers.  This code may be a little
** more picky about the status it sees at various times.  A real
** BIOS probably would not check the status as carefully.
**
** Compile with one of the Borland C or C++ compilers.
**
** This C source contains the main body of the driver code:
** Determine what devices are present, issue ATA Soft Reset,
** execute Non-Data, PIO data in and PIO data out commands.
**
** Note: ATA-4 has added a few timing requirements for the
** host software that were not previously in ATA-2 or ATA-3.
** The new requirements are not implemented in this code
** and are probably never required for anything other than
** ATAPI devices.
*********************************************************************/

#include <dos.h>

#include "ataio.h"

#include "ataiopd.h"

/**************************************************************
**
** extended error infomation,
** config information
** atapi delay flag,
** and a macro for a ~400ns delay.
**
**************************************************************/

struct REG_CMD_INFO reg_cmd_info;

int reg_config_info[2];

int reg_atapi_delay_flag;

long reg_atapi_max_bytes = 32768L;

/*
** ATAPI command packet size and data
*/

int reg_atapi_cp_size;
unsigned char reg_atapi_cp_data[16];

/*
** It is recommended that the host delay 400ns after
** writing the command register.
** Plus...
** ATA-4 has added a new requirement that the host delay
** 400ns if the DEV bit in the Device/Head register is
** changed.  This was not recommended or required in ATA-1,
** ATA-2 or ATA-3.  This is the easy way to do that since it
** works in all PIO modes.
*/

#define DELAY400NS  { pio_inbyte( CB_ASTAT ); pio_inbyte( CB_ASTAT );  \
                      pio_inbyte( CB_ASTAT ); pio_inbyte( CB_ASTAT ); }

/**************************************************************
**
** zero_return_data() -- zero the return data areas.
**
**************************************************************/

static void zero_return_data( void );

static void zero_return_data( void )

{

   reg_cmd_info.flg = TRC_FLAG_EMPTY;
   reg_cmd_info.cmd = 0;
   reg_cmd_info.fr1 = 0;
   reg_cmd_info.sc1 = 0;
   reg_cmd_info.sn1 = 0;
   reg_cmd_info.cl1 = 0;
   reg_cmd_info.ch1 = 0;
   reg_cmd_info.dh1 = 0;
   reg_cmd_info.dc1 = 0;
   reg_cmd_info.ec  = 0;
   reg_cmd_info.to  = 0;
   reg_cmd_info.st2 = 0;
   reg_cmd_info.as2 = 0;
   reg_cmd_info.er2 = 0;
   reg_cmd_info.sc2 = 0;
   reg_cmd_info.sn2 = 0;
   reg_cmd_info.cl2 = 0;
   reg_cmd_info.ch2 = 0;
   reg_cmd_info.dh2 = 0;
   reg_cmd_info.totalBytesXfer = 0L;
   reg_cmd_info.failbits = 0;
}

/**************************************************************
**
** trace_command() -- trace the end of a command.
**
**************************************************************/

static void trace_command( void );

static void trace_command( void )

{

   reg_cmd_info.st2 = pio_inbyte( CB_STAT );
   reg_cmd_info.as2 = pio_inbyte( CB_ASTAT );
   reg_cmd_info.er2 = pio_inbyte( CB_ERR );
   reg_cmd_info.sc2 = pio_inbyte( CB_SC );
   reg_cmd_info.sn2 = pio_inbyte( CB_SN );
   reg_cmd_info.cl2 = pio_inbyte( CB_CL );
   reg_cmd_info.ch2 = pio_inbyte( CB_CH );
   reg_cmd_info.dh2 = pio_inbyte( CB_DH );
   trc_cht();
}

/**************************************************************
**
** reg_atapi_delay() - delay for at least two ticks of the bios
**                     timer (or at least 110ms).
**
**************************************************************/

static void reg_atapi_delay( int dev );

static void reg_atapi_delay( int dev )

{
   int ndx;
   long lw;

   if ( reg_config_info[dev] != REG_CONFIG_TYPE_ATAPI )
      return;
   if ( ! reg_atapi_delay_flag )
      return;
   trc_llt( 0, 0, TRC_LLT_DELAY );
   for ( ndx = 0; ndx < 3; ndx ++ )
   {
      lw = tmr_read_bios_timer();
      while ( lw == tmr_read_bios_timer() )
         /* do nothing */ ;
   }
}

/**************************************************************
**
** reg_wait_poll() - wait for INT 7x or poll for BSY=0
**
**************************************************************/

static void reg_wait_poll( int we, int pe );

static void reg_wait_poll( int we, int pe )

{
   unsigned char status;

   #if 0
   trc_llt( 0, int_intr_flag, TRC_LLT_DEBUG );  /* for debugging */
   #endif

   /*
   ** Wait for INT 7x -or- wait for not BUSY -or- wait for time out.
   */

   if ( int_use_intr_flag )
   {
      trc_llt( 0, 0, TRC_LLT_WINT );
      while ( 1 )
      {
         if ( int_intr_flag )                /* wait for INT 7x */
            break;
         if ( tmr_chk_timeout() )            /* time out yet ? */
         {
            trc_llt( 0, 0, TRC_LLT_TOUT );
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = we;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;
         }
      }
   }
   else
   {
      trc_llt( 0, 0, TRC_LLT_PNBSY );
      while ( 1 )
      {
         status = pio_inbyte( CB_ASTAT );       /* poll for not busy */
         if ( ( status & CB_STAT_BSY ) == 0 )
            break;
         if ( tmr_chk_timeout() )               /* time out yet ? */
         {
            trc_llt( 0, 0, TRC_LLT_TOUT );
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = pe;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;
         }
      }
   }

   #if 0
   trc_llt( 0, int_intr_flag, TRC_LLT_DEBUG );  /* for debugging */
   #endif

   /*
   ** Reset the interrupt flag.
   */

   if ( int_intr_flag > 1 )
      reg_cmd_info.failbits |= FAILBIT15;
   int_intr_flag = 0;
}

/**************************************************************
**
** reg_config() - Check the host adapter and determine the
**                number and type of drives attached.
**
** This process is not documented by any of the ATA standards.
**
** Infomation is returned by this function is in
** reg_config_info[] -- see ATAIO.H.
**
**************************************************************/

int reg_config( void )

{
   int numDev = 0;
   unsigned char sc;
   unsigned char sn;
   unsigned char cl;
   unsigned char ch;
   unsigned char st;
   unsigned char devCtrl;

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );

   /*
   ** mark start of config in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_CFG );

   /*
   ** assume there are no devices
   */

   reg_config_info[0] = REG_CONFIG_TYPE_NONE;
   reg_config_info[1] = REG_CONFIG_TYPE_NONE;

   /*
   ** set up Device Control register
   */

   pio_outbyte( CB_DC, devCtrl );

   /*
   ** lets see if there is a device 0
   */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   pio_outbyte( CB_SC, 0xaa );
   pio_outbyte( CB_SN, 0x55 );
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x55 ) && ( sn == 0xaa ) )
      reg_config_info[0] = REG_CONFIG_TYPE_UNKN;

   /*
   ** lets see if there is a device 1
   */

   pio_outbyte( CB_DH, CB_DH_DEV1 );
   DELAY400NS;
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   pio_outbyte( CB_SC, 0xaa );
   pio_outbyte( CB_SN, 0x55 );
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x55 ) && ( sn == 0xaa ) )
      reg_config_info[1] = REG_CONFIG_TYPE_UNKN;

   /*
   ** now we think we know which devices, if any are there,
   ** so lets try a soft reset (ignoring any errors).
   */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   reg_reset( 0, 0 );

   /*
   ** lets check device 0 again, is device 0 really there?
   ** is it ATA or ATAPI?
   */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
   {
      reg_config_info[0] = REG_CONFIG_TYPE_UNKN;
      cl = pio_inbyte( CB_CL );
      ch = pio_inbyte( CB_CH );
      st = pio_inbyte( CB_STAT );
      if ( ( cl == 0x14 ) && ( ch == 0xeb ) )
         reg_config_info[0] = REG_CONFIG_TYPE_ATAPI;
      else
         if ( ( cl == 0x00 ) && ( ch == 0x00 ) && ( st != 0x00 ) )
            reg_config_info[0] = REG_CONFIG_TYPE_ATA;
   }

   /*
   ** lets check device 1 again, is device 1 really there?
   ** is it ATA or ATAPI?
   */

   pio_outbyte( CB_DH, CB_DH_DEV1 );
   DELAY400NS;
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
   {
      reg_config_info[1] = REG_CONFIG_TYPE_UNKN;
      cl = pio_inbyte( CB_CL );
      ch = pio_inbyte( CB_CH );
      st = pio_inbyte( CB_STAT );
      if ( ( cl == 0x14 ) && ( ch == 0xeb ) )
         reg_config_info[1] = REG_CONFIG_TYPE_ATAPI;
      else
         if ( ( cl == 0x00 ) && ( ch == 0x00 ) && ( st != 0x00 ) )
            reg_config_info[1] = REG_CONFIG_TYPE_ATA;
   }

   /*
   ** If possible, select a device that exists, try device 0 first.
   */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV1 );
      DELAY400NS;
      numDev ++ ;
   }
   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV0 );
      DELAY400NS;
      numDev ++ ;
   }

   /*
   ** mark end of config in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_CFG );

   /*
   ** return the number of devices found
   */

   return numDev;
}

/**************************************************************
**
** reg_reset() - Execute a Software Reset.
**
** See ATA-2 Section 9.2, ATA-3 Section 9.2, ATA-4 Section 8.3.
**
**************************************************************/

int reg_reset( int skipFlag, int devRtrn )

{
   unsigned char sc;
   unsigned char sn;
   unsigned char status;
   unsigned char devCtrl;

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );

   /*
   ** mark start of reset in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_RST );

   /*
   ** Reset error return data.
   */

   zero_return_data();
   reg_cmd_info.flg = TRC_FLAG_SRST;

   /*
   ** initialize the command timeout counter
   */

   tmr_set_timeout();

   /*
   ** Set and then reset the soft reset bit in the Device Control
   ** register.  This causes device 0 be selected.
   */

   if ( ! skipFlag )
   {
      pio_outbyte( CB_DC, devCtrl | CB_DC_SRST );
      DELAY400NS;
      pio_outbyte( CB_DC, devCtrl );
      DELAY400NS;
   }

   /*
   ** If there is a device 0, wait for device 0 to set BSY=0.
   */

   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      reg_atapi_delay( 0 );
      trc_llt( 0, 0, TRC_LLT_PNBSY );
      while ( 1 )
      {
         status = pio_inbyte( CB_STAT );
         if ( ( status & CB_STAT_BSY ) == 0 )
            break;
         if ( tmr_chk_timeout() )
         {
            trc_llt( 0, 0, TRC_LLT_TOUT );
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 1;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;
         }
      }
   }

   /*
   ** If there is a device 1, wait until device 1 allows
   ** register access.
   */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      reg_atapi_delay( 1 );
      trc_llt( 0, 0, TRC_LLT_PNBSY );
      while ( 1 )
      {
         pio_outbyte( CB_DH, CB_DH_DEV1 );
         DELAY400NS;
         sc = pio_inbyte( CB_SC );
         sn = pio_inbyte( CB_SN );
         if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
            break;
         if ( tmr_chk_timeout() )
         {
            trc_llt( 0, 0, TRC_LLT_TOUT );
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 2;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;
         }
      }

      /*
      ** Now check if drive 1 set BSY=0.
      */

      if ( reg_cmd_info.ec == 0 )
      {
         if ( pio_inbyte( CB_STAT ) & CB_STAT_BSY )
         {
            reg_cmd_info.ec = 3;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         }
      }
   }

   /*
   ** RESET_DONE:
   */

   /*
   ** We are done but now we must select the device the caller
   ** requested before we trace the command.  This will cause
   ** the correct data to be returned in reg_cmd_info.
   */

   pio_outbyte( CB_DH, devRtrn ? CB_DH_DEV1 : CB_DH_DEV0 );
   DELAY400NS;
   trace_command();

   /*
   ** If possible, select a device that exists,
   ** try device 0 first.
   */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV1 );
      DELAY400NS;
   }
   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV0 );
      DELAY400NS;
   }

   /*
   ** mark end of reset in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_RST );

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/**************************************************************
**
** reg_select() - function used to select a drive.
**
** Function to select a drive. This subroutine waits for not BUSY,
** selects a drive and waits for READY and SEEK COMPLETE status.
**
***************************************************************/

static int reg_select( int dev );

static int reg_select( int dev )

{
   unsigned char status;

   /*
   ** PAY ATTENTION HERE
   ** The caller may want to issue a command to a device that doesn't
   ** exist (for example, Exec Dev Diag), so if we see this,
   ** just select that device, skip all status checking and return.
   ** We assume the caller knows what they are doing!
   */

   if ( reg_config_info[ dev ] < REG_CONFIG_TYPE_ATA )
   {
      /*
      ** select the device and return
      */

      pio_outbyte( CB_DH, dev ? CB_DH_DEV1 : CB_DH_DEV0 );
      DELAY400NS;
      return 0;
   }

   /*
   ** The rest of this is the normal ATA stuff for device selection
   ** and we don't expect the caller to be selecting a device that
   ** does not exist.
   ** We don't know which drive is currently selected but we should
   ** wait for it to be not BUSY.  Normally it will be not BUSY
   ** unless something is very wrong!
   */

   trc_llt( 0, 0, TRC_LLT_PNBSY );
   while ( 1 )
   {
      status = pio_inbyte( CB_STAT );
      if ( ( status & CB_STAT_BSY ) == 0 )
         break;
      if ( tmr_chk_timeout() )
      {
         trc_llt( 0, 0, TRC_LLT_TOUT );
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 11;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         reg_cmd_info.st2 = status;
         reg_cmd_info.as2 = pio_inbyte( CB_ASTAT );
         reg_cmd_info.er2 = pio_inbyte( CB_ERR );
         reg_cmd_info.sc2 = pio_inbyte( CB_SC );
         reg_cmd_info.sn2 = pio_inbyte( CB_SN );
         reg_cmd_info.cl2 = pio_inbyte( CB_CL );
         reg_cmd_info.ch2 = pio_inbyte( CB_CH );
         reg_cmd_info.dh2 = pio_inbyte( CB_DH );
         return 1;
      }
   }

   /*
   ** Here we select the drive we really want to work with by
   ** putting 0xA0 or 0xB0 in the Drive/Head register (1f6).
   */

   pio_outbyte( CB_DH, dev ? CB_DH_DEV1 : CB_DH_DEV0 );
   DELAY400NS;

   /*
   ** If the selected device is an ATA device,
   ** wait for it to have READY and SEEK COMPLETE
   ** status.  Normally the drive should be in this state unless
   ** something is very wrong (or initial power up is still in
   ** progress).  For any other type of device, just wait for
   ** BSY=0 and assume the caller knows what they are doing.
   */

   trc_llt( 0, 0, TRC_LLT_PRDY );
   while ( 1 )
   {
      status = pio_inbyte( CB_STAT );
      if ( reg_config_info[ dev ] == REG_CONFIG_TYPE_ATA )
      {
           if ( ( status & ( CB_STAT_BSY | CB_STAT_RDY | CB_STAT_SKC ) )
                     == ( CB_STAT_RDY | CB_STAT_SKC ) )
         break;
      }
      else
      {
         if ( ( status & CB_STAT_BSY ) == 0 )
            break;
      }
      if ( tmr_chk_timeout() )
      {
         trc_llt( 0, 0, TRC_LLT_TOUT );
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 12;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         reg_cmd_info.st2 = status;
         reg_cmd_info.as2 = pio_inbyte( CB_ASTAT );
         reg_cmd_info.er2 = pio_inbyte( CB_ERR );
         reg_cmd_info.sc2 = pio_inbyte( CB_SC );
         reg_cmd_info.sn2 = pio_inbyte( CB_SN );
         reg_cmd_info.cl2 = pio_inbyte( CB_CL );
         reg_cmd_info.ch2 = pio_inbyte( CB_CH );
         reg_cmd_info.dh2 = pio_inbyte( CB_DH );
         return 1;
      }
   }

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/**************************************************************
**
** reg_non_data_lba() - Easy way to execute a non-data command
**                      using an LBA sector address.
**
**************************************************************/

int reg_non_data_lba( int dev, int cmd,
                      int fr, int sc,
                      long lba )

{
   unsigned int cyl;
   int head, sect;

   sect = (int) ( lba & 0x000000ffL );
   lba = lba >> 8;
   cyl = (int) ( lba & 0x0000ffff );
   lba = lba >> 16;
   head = ( (int) ( lba & 0x0fL ) ) | 0x40;
   return reg_non_data( dev, cmd,
                        fr, sc,
                        cyl, head, sect );
}

/**************************************************************
**
** reg_non_data() - Execute a non-data command.
**
** Note special handling for Execute Device Diagnostics
** command when there is no device 0.
**
** See ATA-2 Section 9.5, ATA-3 Section 9.5,
** ATA-4 Section 8.8 Figure 12.  Also see Section 8.5.
**
**************************************************************/

int reg_non_data( int dev, int cmd,
                   int fr, int sc,
                   unsigned int cyl, int head, int sect )

{
   unsigned char secCnt;
   unsigned char secNum;
   unsigned char devHead;
   unsigned char devCtrl;
   unsigned char cylLow;
   unsigned char cylHigh;
   unsigned char status;

   /*
   ** mark start of ND cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_ND );

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   devHead = dev ? CB_DH_DEV1 : CB_DH_DEV0;
   devHead = devHead | ( head & 0x4f );
   cylLow = cyl & 0x00ff;
   cylHigh = ( cyl & 0xff00 ) >> 8;

   /*
   ** Reset error return data.
   */

   zero_return_data();
   reg_cmd_info.flg = TRC_FLAG_CMD;
   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr1 = fr;
   reg_cmd_info.sc1 = sc;
   reg_cmd_info.sn1 = sect;
   reg_cmd_info.cl1 = cylLow;
   reg_cmd_info.ch1 = cylHigh;
   reg_cmd_info.dh1 = devHead;
   reg_cmd_info.dc1 = devCtrl;

   /*
   ** Set command time out.
   */

   tmr_set_timeout();

   /*
   ** Select the drive - call the reg_select function (above).
   ** Quit now if this fails.
   */

   if ( reg_select( dev ) )
   {
      trace_command();
      trc_llt( 0, 0, TRC_LLT_E_ND );
      return 1;
   }

   /*
   ** Set up all the registers except the command register.
   */

   pio_outbyte( CB_DC, devCtrl );
   pio_outbyte( CB_FR, fr );
   pio_outbyte( CB_SC, sc );
   pio_outbyte( CB_SN, sect );
   pio_outbyte( CB_CL, cylLow );
   pio_outbyte( CB_CH, cylHigh );
   pio_outbyte( CB_DH, devHead );

   /*
   ** For interrupt mode,
   ** Take over INT 7x and initialize interrupt controller
   ** and reset interrupt flag.
   */

   int_save_int_vect();

   /*
   ** Start the command by setting the Command register.  The drive
   ** should immediately set BUSY status.
   */

   pio_outbyte( CB_CMD, cmd );

   /*
   ** Waste some time by reading the alternate status a few times.
   ** This gives the drive time to set BUSY in the status register on
   ** really fast systems.  If we don't do this, a slow drive on a fast
   ** system may not set BUSY fast enough and we would think it had
   ** completed the command when it really had not even started the
   ** command yet.
   */

   DELAY400NS;

   if ( reg_config_info[0] == REG_CONFIG_TYPE_ATAPI )
      reg_atapi_delay( 0 );
   if ( reg_config_info[1] == REG_CONFIG_TYPE_ATAPI )
      reg_atapi_delay( 1 );

   /*
   ** IF
   ** This is an Exec Dev Diag command (cmd=0x90)
   ** and there is no device 0, then
   ** there will be no interrupt. So we must
   ** poll device 1 until it allows register
   ** access and then do normal polling of the Status
   ** register for BSY=0.
   ** ELSE
   ** Do the normal wait for interrupt or polling for
   ** completion.
   */

   if ( ( cmd == CMD_EXECUTE_DEVICE_DIAGNOSTIC )
        &&
        ( reg_config_info[0] == REG_CONFIG_TYPE_NONE )
      )
   {
      trc_llt( 0, 0, TRC_LLT_PNBSY );
      while ( 1 )
      {
         pio_outbyte( CB_DH, CB_DH_DEV1 );
         DELAY400NS;
         secCnt = pio_inbyte( CB_SC );
         secNum = pio_inbyte( CB_SN );
         if ( ( secCnt == 0x01 ) && ( secNum == 0x01 ) )
            break;
         if ( tmr_chk_timeout() )
         {
            trc_llt( 0, 0, TRC_LLT_TOUT );
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 24;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;
         }
      }
   }
   else
   {
      /*
      ** Wait for INT 7x -or- wait for not BUSY -or- wait for time out.
      */

      reg_wait_poll( 22, 23 );
   }

   /*
   ** Read the primary status register.  In keeping with the rules
   ** stated above the primary status register is read only
   ** ONCE.
   */

   status = pio_inbyte( CB_STAT );

   /*
   ** Error if BUSY, WRITE FAULT, DRQ or ERROR status now.
   */

   if ( reg_cmd_info.ec == 0 )
   {
      if ( status & ( CB_STAT_BSY | CB_STAT_WFT | CB_STAT_DRQ | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 21;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
      }
   }

   /*
   ** read the output registers and trace the command.
   */

   trace_command();

   /*
   ** NON_DATA_DONE:
   */

   /*
   ** For interrupt mode, restore the INT 7x vector.
   */

   int_restore_int_vect();

   /*
   ** mark end of ND cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_ND );

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/**************************************************************
**
** reg_pio_data_in_lba() - Easy way to execute a PIO Data In command
**                         using an LBA sector address.
**
**************************************************************/

int reg_pio_data_in_lba( int dev, int cmd,
                         int fr, int sc,
                         long lba,
                         unsigned seg, unsigned off,
                         int numSect, int multiCnt )

{
   unsigned int cyl;
   int head, sect;

   sect = (int) ( lba & 0x000000ffL );
   lba = lba >> 8;
   cyl = (int) ( lba & 0x0000ffff );
   lba = lba >> 16;
   head = ( (int) ( lba & 0x0fL ) ) | 0x40;
   return reg_pio_data_in( dev, cmd,
                           fr, sc,
                           cyl, head, sect,
                           seg, off,
                           numSect, multiCnt );
}

/**************************************************************
**
** reg_pio_data_in() - Execute a PIO Data In command.
**
** See ATA-2 Section 9.3, ATA-3 Section 9.3,
** ATA-4 Section 8.6 Figure 10.
**
**************************************************************/

int reg_pio_data_in( int dev, int cmd,
                     int fr, int sc,
                     unsigned int cyl, int head, int sect,
                     unsigned seg, unsigned off,
                     int numSect, int multiCnt )

{
   unsigned char devHead;
   unsigned char devCtrl;
   unsigned char cylLow;
   unsigned char cylHigh;
   unsigned char status;
   unsigned int wordCnt;

   /*
   ** mark start of PDI cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_PDI );

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   devHead = dev ? CB_DH_DEV1 : CB_DH_DEV0;
   devHead = devHead | ( head & 0x4f );
   cylLow = cyl & 0x00ff;
   cylHigh = ( cyl & 0xff00 ) >> 8;

   /*
   ** Reset error return data.
   */

   zero_return_data();
   reg_cmd_info.flg = TRC_FLAG_CMD;
   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr1 = fr;
   reg_cmd_info.sc1 = sc;
   reg_cmd_info.sn1 = sect;
   reg_cmd_info.cl1 = cylLow;
   reg_cmd_info.ch1 = cylHigh;
   reg_cmd_info.dh1 = devHead;
   reg_cmd_info.dc1 = devCtrl;

   /*
   ** Set command time out.
   */

   tmr_set_timeout();

   /*
   ** Select the drive - call the reg_select function (above).
   ** Quit now if this fails.
   */

   if ( reg_select( dev ) )
   {
      trace_command();
      trc_llt( 0, 0, TRC_LLT_E_PDI );
      return 1;
   }

   /*
   ** Set up all the registers except the command register.
   */

   pio_outbyte( CB_DC, devCtrl );
   pio_outbyte( CB_FR, fr );
   pio_outbyte( CB_SC, sc );
   pio_outbyte( CB_SN, sect );
   pio_outbyte( CB_CL, cylLow );
   pio_outbyte( CB_CH, cylHigh );
   pio_outbyte( CB_DH, devHead );

   /*
   ** For interrupt mode,
   ** Take over INT 7x and initialize interrupt controller
   ** and reset interrupt flag.
   */

   int_save_int_vect();

   /*
   ** Start the command by setting the Command register.  The drive
   ** should immediately set BUSY status.
   */

   pio_outbyte( CB_CMD, cmd );

   /*
   ** Waste some time by reading the alternate status a few times.
   ** This gives the drive time to set BUSY in the status register on
   ** really fast systems.  If we don't do this, a slow drive on a fast
   ** system may not set BUSY fast enough and we would think it had
   ** completed the command when it really had not even started the
   ** command yet.
   */

   DELAY400NS;

   /*
   ** Loop to read each sector.
   */

   while ( 1 )
   {
      /*
      ** READ_LOOP:
      **
      ** NOTE NOTE NOTE ...  The primary status register (1f7) MUST NOT be
      ** read more than ONCE for each sector transferred!  When the
      ** primary status register is read, the drive resets IRQ 14.  The
      ** alternate status register (3f6) can be read any number of times.
      ** After INT 7x read the the primary status register ONCE
      ** and transfer the 256 words (REP INSW).  AS SOON as BOTH the
      ** primary status register has been read AND the last of the 256
      ** words has been read, the drive is allowed to generate the next
      ** IRQ 14 (newer and faster drives could generate the next IRQ 14 in
      ** 50 microseconds or less).  If the primary status register is read
      ** more than once, there is the possibility of a race between the
      ** drive and the software and the next IRQ 14 could be reset before
      ** the system interrupt controller sees it.
      */

      /*
      ** Wait for INT 7x -or- wait for not BUSY -or- wait for time out.
      */

      reg_atapi_delay( dev );
      reg_wait_poll( 34, 35 );

      /*
      ** Read the primary status register.  In keeping with the rules
      ** stated above the primary status register is read only
      ** ONCE.
      */

      status = pio_inbyte( CB_STAT );

      /*
      ** If there was a time out error, go to READ_DONE.
      */

      if ( reg_cmd_info.ec )
         break;   /* go to READ_DONE */

      /*
      ** If BSY=0 and DRQ=1, transfer the data,
      ** even if we find out there is an error later.
      */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == CB_STAT_DRQ )
      {
         /*
         ** determine the number of sectors to transfer
         */

         wordCnt = multiCnt ? multiCnt : 1;
         if ( wordCnt > numSect )
            wordCnt = numSect;
         wordCnt = wordCnt * 256;

         /*
         ** Do the REP INSW to read the data for one block.
         */

         reg_cmd_info.totalBytesXfer += ( wordCnt << 1 );
         pio_rep_inword( CB_DATA, seg, off, wordCnt );

         /*
         ** Note: The drive should have dropped DATA REQUEST by now.  If there
         ** are more sectors to transfer, BUSY should be active now (unless
         ** there is an error).
         */

         /*
         ** Decrement the count of sectors to be transferred
         ** and increment buffer address.
         */

         numSect = numSect - ( multiCnt ? multiCnt : 1 );
         seg = seg + ( 32 * ( multiCnt ? multiCnt : 1 ) );
      }

      /*
      ** So was there any error condition?
      */

      if ( status & ( CB_STAT_BSY | CB_STAT_WFT | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 31;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         break;   /* go to READ_DONE */
      }

      /*
      ** DRQ should have been set -- was it?
      */

      if ( ( status & CB_STAT_DRQ ) == 0 )
      {
         reg_cmd_info.ec = 32;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         break;   /* go to READ_DONE */
      }

      /*
      ** If all of the requested sectors have been transferred, make a
      ** few more checks before we exit.
      */

      if ( numSect < 1 )
      {
         /*
         ** Since the drive has transferred all of the requested sectors
         ** without error, the drive should not have BUSY, WRITE FAULT,
         ** DATA REQUEST or ERROR active now.
         */

         status = pio_inbyte( CB_STAT );
         if ( status & ( CB_STAT_BSY | CB_STAT_WFT | CB_STAT_DRQ | CB_STAT_ERR ) )
         {
            reg_cmd_info.ec = 33;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;   /* go to READ_DONE */
         }

         /*
         ** All sectors have been read without error, go to READ_DONE.
         */

         break;   /* go to READ_DONE */

      }

      /*
      ** This is the end of the read loop.  If we get here, the loop is
      ** repeated to read the next sector.  Go back to READ_LOOP.
      */

   }

   /*
   ** read the output registers and trace the command.
   */

   trace_command();

   /*
   ** READ_DONE:
   */

   /*
   ** For interrupt mode, restore the INT 7x vector.
   */

   int_restore_int_vect();

   /*
   ** mark end of PDI cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_PDI );

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/**************************************************************
**
** reg_pio_data_out_lba() - Easy way to execute a PIO Data In command
**                          using an LBA sector address.
**
**************************************************************/

int reg_pio_data_out_lba( int dev, int cmd,
                          int fr, int sc,
                          long lba,
                          unsigned seg, unsigned off,
                          int numSect, int multiCnt )

{
   unsigned int cyl;
   int head, sect;

   sect = (int) ( lba & 0x000000ffL );
   lba = lba >> 8;
   cyl = (int) ( lba & 0x0000ffff );
   lba = lba >> 16;
   head = ( (int) ( lba & 0x0fL ) ) | 0x40;
   return reg_pio_data_out( dev, cmd,
                            fr, sc,
                            cyl, head, sect,
                            seg, off,
                            numSect, multiCnt );
}

/**************************************************************
**
** reg_pio_data_out() - Execute a PIO Data Out command.
**
** See ATA-2 Section 9.4, ATA-3 Section 9.4,
** ATA-4 Section 8.7 Figure 11.
**
**************************************************************/

int reg_pio_data_out( int dev, int cmd,
                      int fr, int sc,
                      unsigned int cyl, int head, int sect,
                      unsigned seg, unsigned off,
                      int numSect, int multiCnt )

{
   unsigned char devHead;
   unsigned char devCtrl;
   unsigned char cylLow;
   unsigned char cylHigh;
   unsigned char status;
   int loopFlag = 1;
   unsigned int wordCnt;

   /*
   ** mark start of PDO cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_PDO );

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   devHead = dev ? CB_DH_DEV1 : CB_DH_DEV0;
   devHead = devHead | ( head & 0x4f );
   cylLow = cyl & 0x00ff;
   cylHigh = ( cyl & 0xff00 ) >> 8;

   /*
   ** Reset error return data.
   */

   zero_return_data();
   reg_cmd_info.flg = TRC_FLAG_CMD;
   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr1 = fr;
   reg_cmd_info.sc1 = sc;
   reg_cmd_info.sn1 = sect;
   reg_cmd_info.cl1 = cylLow;
   reg_cmd_info.ch1 = cylHigh;
   reg_cmd_info.dh1 = devHead;
   reg_cmd_info.dc1 = devCtrl;

   /*
   ** Set command time out.
   */

   tmr_set_timeout();

   /*
   ** Select the drive - call the reg_select function (above).
   ** Quit now if this fails.
   */

   if ( reg_select( dev ) )
   {
      trace_command();
      trc_llt( 0, 0, TRC_LLT_E_PDO );
      return 1;
   }

   /*
   ** Set up all the registers except the command register.
   */

   pio_outbyte( CB_DC, devCtrl );
   pio_outbyte( CB_FR, fr );
   pio_outbyte( CB_SC, sc );
   pio_outbyte( CB_SN, sect );
   pio_outbyte( CB_CL, cylLow );
   pio_outbyte( CB_CH, cylHigh );
   pio_outbyte( CB_DH, devHead );

   /*
   ** For interrupt mode,
   ** Take over INT 7x and initialize interrupt controller
   ** and reset interrupt flag.
   */

   int_save_int_vect();

   /*
   ** Start the command by setting the Command register.  The drive
   ** should immediately set BUSY status.
   */

   pio_outbyte( CB_CMD, cmd );

   /*
   ** Waste some time by reading the alternate status a few times.
   ** This gives the drive time to set BUSY in the status register on
   ** really fast systems.  If we don't do this, a slow drive on a fast
   ** system may not set BUSY fast enough and we would think it had
   ** completed the command when it really had not even started the
   ** command yet.
   */

   DELAY400NS;

   /*
   ** Wait for not BUSY or time out.
   ** Note: No interrupt is generated for the
   ** first sector of a write command.  Well...
   ** that's not really true we are working with
   ** a PCMCIA PC Card ATA device.
   */

   reg_atapi_delay( dev );
   trc_llt( 0, 0, TRC_LLT_PNBSY );
   while ( 1 )
   {
      status = pio_inbyte( CB_ASTAT );
      if ( ( status & CB_STAT_BSY ) == 0 )
         break;
      if ( tmr_chk_timeout() )
      {
         trc_llt( 0, 0, TRC_LLT_TOUT );
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 47;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         loopFlag = 0;
         break;
      }
   }

   /*
   ** If we are using interrupts and we just got an interrupt, this is
   ** wrong.  The drive must not generate an interrupt at this time.
   */

   if ( loopFlag && int_use_intr_flag && int_intr_flag )
   {
      reg_cmd_info.ec = 46;
      trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
      loopFlag = 0;
   }

   /*
   ** This loop writes each sector.
   */

   while ( loopFlag )
   {
      /*
      ** WRITE_LOOP:
      **
      ** NOTE NOTE NOTE ...  The primary status register (1f7) MUST NOT be
      ** read more than ONCE for each sector transferred!  When the
      ** primary status register is read, the drive resets IRQ 14.  The
      ** alternate status register (3f6) can be read any number of times.
      ** For correct results, transfer the 256 words (REP OUTSW), wait for
      ** INT 7x and then read the primary status register.  AS
      ** SOON as BOTH the primary status register has been read AND the
      ** last of the 256 words has been written, the drive is allowed to
      ** generate the next IRQ 14 (newer and faster drives could generate
      ** the next IRQ 14 in 50 microseconds or less).  If the primary
      ** status register is read more than once, there is the possibility
      ** of a race between the drive and the software and the next IRQ 14
      ** could be reset before the system interrupt controller sees it.
      */

      /*
      ** If BSY=0 and DRQ=1, transfer the data,
      ** even if we find out there is an error later.
      */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == CB_STAT_DRQ )
      {
         /*
         ** determine the number of sectors to transfer
         */

         wordCnt = multiCnt ? multiCnt : 1;
         if ( wordCnt > numSect )
            wordCnt = numSect;
         wordCnt = wordCnt * 256;

         /*
         ** Do the REP OUTSW to write the data for one block.
         */

         reg_cmd_info.totalBytesXfer += ( wordCnt << 1 );
         pio_rep_outword( CB_DATA, seg, off, wordCnt );

         /*
         ** Note: The drive should have dropped DATA REQUEST and
         ** raised BUSY by now.
         */

         /*
         ** Decrement the count of sectors to be transferred
         ** and increment buffer address.
         */

         numSect = numSect - ( multiCnt ? multiCnt : 1 );
         seg = seg + ( 32 * ( multiCnt ? multiCnt : 1 ) );
      }

      /*
      ** So was there any error condition?
      */

      if ( status & ( CB_STAT_BSY | CB_STAT_WFT | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 41;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         break;   /* go to WRITE_DONE */
      }

      /*
      ** DRQ should have been set -- was it?
      */

      if ( ( status & CB_STAT_DRQ ) == 0 )
      {
         reg_cmd_info.ec = 42;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         break;   /* go to WRITE_DONE */
      }

      /*
      ** Wait for INT 7x -or- wait for not BUSY -or- wait for time out.
      */

      reg_atapi_delay( dev );
      reg_wait_poll( 44, 45 );

      /*
      ** Read the primary status register.  In keeping with the rules
      ** stated above the primary status register is read only ONCE.
      */

      status = pio_inbyte( CB_STAT );

      /*
      ** If there was a time out error, go to WRITE_DONE.
      */

      if ( reg_cmd_info.ec )
         break;   /* go to WRITE_DONE */

      /*
      ** If all of the requested sectors have been transferred, make a
      ** few more checks before we exit.
      */

      if ( numSect < 1 )
      {
         /*
         ** Since the drive has transferred all of the sectors without
         ** error, the drive MUST not have BUSY, WRITE FAULT, DATA REQUEST
         ** or ERROR status at this time.
         */

         if ( status & ( CB_STAT_BSY | CB_STAT_WFT | CB_STAT_DRQ | CB_STAT_ERR ) )
         {
            reg_cmd_info.ec = 43;
            trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
            break;   /* go to WRITE_DONE */
         }

         /*
         ** All sectors have been written without error, go to WRITE_DONE.
         */

         break;   /* go to WRITE_DONE */

      }

      /*
      **
      ** This is the end of the write loop.  If we get here, the loop
      ** is repeated to write the next sector.  Go back to WRITE_LOOP.
      */

   }

   /*
   ** read the output registers and trace the command.
   */

   trace_command();

   /*
   ** WRITE_DONE:
   */

   /*
   ** For interrupt mode, restore the INT 7x vector.
   */

   int_restore_int_vect();

   /*
   ** mark end of PDO cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_PDO );

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/**************************************************************
**
** reg_packet() - Execute an ATAPI Packet (A0H) command.
**
** See ATA-4 Section 9.10, Figure 14.
**
**************************************************************/

int reg_packet( int dev,
                unsigned int cpbc,
                unsigned int pseg, unsigned int poff,
                int dir,
                unsigned int dpbc,
                unsigned int dseg, unsigned int doff )

{
   unsigned char devCtrl;
   unsigned char devHead;
   unsigned char cylLow;
   unsigned char cylHigh;
   unsigned char status;
   unsigned char reason;
   unsigned char lowCyl;
   unsigned char highCyl;
   unsigned int byteCnt;
   unsigned int wordCnt;
   int ndx;
   unsigned char * cp;
   unsigned char far * cfp;

   /*
   ** mark start of PI cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_S_PI );

   /*
   ** setup register values
   */

   devCtrl = CB_DC_HD15 | ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   devHead = dev ? CB_DH_DEV1 : CB_DH_DEV0;
   cylLow = dpbc & 0x00ff;
   cylHigh = ( dpbc & 0xff00 ) >> 8;

   /*
   ** Reset error return data.
   */

   zero_return_data();
   reg_cmd_info.flg = TRC_FLAG_CMD;
   reg_cmd_info.cmd = CMD_PACKET;
   reg_cmd_info.fr1 = 0;
   reg_cmd_info.sc1 = 0;
   reg_cmd_info.sn1 = 0;
   reg_cmd_info.cl1 = cylLow;
   reg_cmd_info.ch1 = cylHigh;
   reg_cmd_info.dh1 = devHead;
   reg_cmd_info.dc1 = devCtrl;

   /*
   ** Make sure the command packet size is either 12 or 16
   ** and save the command packet size and data.
   */

   cpbc = cpbc < 12 ? 12 : cpbc;
   cpbc = cpbc > 12 ? 16 : cpbc;
   reg_atapi_cp_size = cpbc;
   cp = reg_atapi_cp_data;
   cfp = MK_FP( pseg, poff );
   for ( ndx = 0; ndx < cpbc; ndx ++ )
   {
      * cp = * cfp;
      cp ++ ;
      cfp ++ ;
   }

   /*
   ** Set command time out.
   */

   tmr_set_timeout();

   /*
   ** Select the drive - call the reg_select function (above).
   ** Quit now if this fails.
   */

   if ( reg_select( dev ) )
   {
      trace_command();
      trc_llt( 0, 0, TRC_LLT_E_PI );
      reg_atapi_max_bytes = 32768L;    /* reset max bytes */
      return 1;
   }

   /*
   ** Set up all the registers except the command register.
   */

   pio_outbyte( CB_DC, devCtrl );
   pio_outbyte( CB_FR, 0 );
   pio_outbyte( CB_SC, 0 );
   pio_outbyte( CB_SN, 0 );
   pio_outbyte( CB_CL, cylLow );
   pio_outbyte( CB_CH, cylHigh );
   pio_outbyte( CB_DH, devHead );

   /*
   ** For interrupt mode,
   ** Take over INT 7x and initialize interrupt controller
   ** and reset interrupt flag.
   */

   int_save_int_vect();

   /*
   ** Start the command by setting the Command register.  The drive
   ** should immediately set BUSY status.
   */

   pio_outbyte( CB_CMD, CMD_PACKET );

   /*
   ** Waste some time by reading the alternate status a few times.
   ** This gives the drive time to set BUSY in the status register on
   ** really fast systems.  If we don't do this, a slow drive on a fast
   ** system may not set BUSY fast enough and we would think it had
   ** completed the command when it really had not even started the
   ** command yet.
   */

   DELAY400NS;

   /*
   ** Command packet transfer...
   ** Check for protocol failures,
   ** the device should have BSY=1 now but any number of
   ** strange things can happen on ATAPI devices:
   ** 1) the device may not set BSY=1 or DRQ=1 for awhile.
   ** 2) the device may go directly to DRQ=1 status,
   ** 3) the device may reject the command (maybe it is
   **    not really an ATAPI device or has some error).
   */

   status = pio_inbyte( CB_ASTAT );
   while ( 1 )
   {
      if ( status & CB_STAT_BSY )
         break;                              /* BSY=1 (that's OK) */
      else
      {
         if ( status & CB_STAT_DRQ )
            break;                           /* BSY=0, DRQ=1 (that's OK) */
         if ( status & CB_STAT_ERR )
            break;                           /* BSY=0, error (that's OK) */
         reg_cmd_info.failbits |= FAILBIT0;   /* not OK */
      }
      if ( tmr_chk_timeout() )               /* time out yet ? */
         break;
      status = pio_inbyte( CB_ASTAT );
   }

   /*
   ** Command packet transfer...
   ** Poll Alternate Status for BSY=0.
   */

   trc_llt( 0, 0, TRC_LLT_PNBSY );
   while ( 1 )
   {
      status = pio_inbyte( CB_ASTAT );       /* poll for not busy */
      if ( ( status & CB_STAT_BSY ) == 0 )
         break;
      if ( tmr_chk_timeout() )               /* time out yet ? */
      {
         trc_llt( 0, 0, TRC_LLT_TOUT );
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 51;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         dir = -1;   /* command done */
         break;
      }
   }

   /*
   ** Command packet transfer...
   ** Check for protocol failures... no interrupt here please!
   ** Clear any interrupt the command packet transfer may have caused.
   */

   if ( int_intr_flag )
      reg_cmd_info.failbits |= FAILBIT1;
   int_intr_flag = 0;

   /*
   ** Command packet transfer...
   ** If no error, transfer the command packet.
   */

   if ( reg_cmd_info.ec == 0 )
   {

      /*
      ** Command packet transfer...
      ** Read the primary status register and the other ATAPI registers.
      */

      status = pio_inbyte( CB_STAT );
      reason = pio_inbyte( CB_SC );
      lowCyl = pio_inbyte( CB_CL );
      highCyl = pio_inbyte( CB_CH );

      /*
      ** Command packet transfer...
      ** check status: must have BSY=0, DRQ=1 now
      */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) != CB_STAT_DRQ )
      {
         reg_cmd_info.ec = 52;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         dir = -1;   /* command done */
      }
      else
      {
         /*
         ** Command packet transfer...
         ** Check for protocol failures...
         ** check: C/nD=1, IO=0.
         */

         if ( ( reason &  ( CB_SC_P_TAG | CB_SC_P_REL | CB_SC_P_IO ) )
              || ( ! ( reason &  CB_SC_P_CD ) )
            )
            reg_cmd_info.failbits |= FAILBIT2;
         if ( ( lowCyl != cylLow ) || ( highCyl != cylHigh ) )
            reg_cmd_info.failbits |= FAILBIT3;

         /*
         ** Command packet transfer...
         ** trace cdb byte 0,
         ** xfer the command packet (the cdb)
         */

         trc_llt( 0, * (unsigned char far *) MK_FP( pseg, poff ), TRC_LLT_P_CMD );
         pio_rep_outword( CB_DATA, pseg, poff, cpbc >> 1 );
      }
   }

   /*
   ** Data transfer loop...
   ** If there is no error, enter the data transfer loop.
   */

   while ( reg_cmd_info.ec == 0 )
   {
      /*
      ** Data transfer loop...
      ** Wait for INT 7x -or- wait for not BUSY -or- wait for time out.
      */

      reg_atapi_delay( dev );
      reg_wait_poll( 53, 54 );

      /*
      ** Data transfer loop...
      ** If there was a time out error, exit the data transfer loop.
      */

      if ( reg_cmd_info.ec )
      {
         dir = -1;   /* command done */
         break;
      }

      /*
      ** Data transfer loop...
      ** Read the primary status register.
      */

      status = pio_inbyte( CB_STAT );
      reason = pio_inbyte( CB_SC );
      lowCyl = pio_inbyte( CB_CL );
      highCyl = pio_inbyte( CB_CH );

      /*
      ** Data transfer loop...
      ** Exit the read data loop if the device indicates this
      ** is the end of the command.
      */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == 0 )
      {
         dir = -1;   /* command done */
         break;
      }

      /*
      ** Data transfer loop...
      ** The device must want to transfer data...
      ** check status: must have BSY=0, DRQ=1 now.
      */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) != CB_STAT_DRQ )
      {
         reg_cmd_info.ec = 55;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         dir = -1;   /* command done */
         break;
      }
      else
      {
         /*
         ** Data transfer loop...
         ** Check for protocol failures...
         ** check: C/nD=0, IO=1 (read) or IO=0 (write).
         */

         if ( ( reason &  ( CB_SC_P_TAG | CB_SC_P_REL ) )
              || ( reason &  CB_SC_P_CD )
            )
            reg_cmd_info.failbits |= FAILBIT4;
         if ( ( reason & CB_SC_P_IO ) && dir )
            reg_cmd_info.failbits |= FAILBIT5;
      }

      /*
      ** Data transfer loop...
      ** get the byte count
      */

      byteCnt = ( highCyl << 8 ) | lowCyl;
      if ( ( byteCnt < 1 ) || ( byteCnt > dpbc ) )
         reg_cmd_info.failbits |= FAILBIT6;
      if ( byteCnt & 0x0001 )
         reg_cmd_info.failbits |= FAILBIT7;

      /*
      ** Data transfer loop...
      ** make sure we don't overrun the caller's buffer
      */

      if ( ( reg_cmd_info.totalBytesXfer + byteCnt ) > reg_atapi_max_bytes )
      {
         reg_cmd_info.ec = 59;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
         dir = -1;   /* command done */
         break;
      }

      /*
      ** Data transfer loop...
      ** transfer the data
      */

      wordCnt = ( byteCnt >> 1 ) + ( byteCnt & 0x0001 );
      reg_cmd_info.totalBytesXfer += ( wordCnt << 1 );
      if ( dir )
         pio_rep_outword( CB_DATA, dseg, doff, wordCnt );
      else
         pio_rep_inword( CB_DATA, dseg, doff, wordCnt );
      doff = doff + byteCnt;              /*??? segment crossing */
   }

   /*
   ** End of command...
   ** Wait for interrupt or poll for BSY=0,
   ** but don't do this if there was any error or if this
   ** was a commmand that did not transfer data.
   */

   if ( ( reg_cmd_info.ec == 0 ) && ( dir >= 0 ) )
   {
      reg_atapi_delay( dev );
      reg_wait_poll( 56, 57 );
   }

   /*
   ** Final status check, only if no previous error.
   */

   if ( reg_cmd_info.ec == 0 )
   {
      /*
      ** Final status check...
      ** Read the primary status register and other regs.
      */

      status = pio_inbyte( CB_STAT );
      reason = pio_inbyte( CB_SC );
      lowCyl = pio_inbyte( CB_CL );
      highCyl = pio_inbyte( CB_CH );

      /*
      ** Final status check...
      ** check for any error.
      */

      if ( status & ( CB_STAT_BSY | CB_STAT_DRQ | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 58;
         trc_llt( 0, reg_cmd_info.ec, TRC_LLT_ERROR );
      }
   }

   /*
   ** Final status check...
   ** Check for protocol failures...
   ** check: C/nD=1, IO=1.
   */

   if ( ( reason & ( CB_SC_P_TAG | CB_SC_P_REL ) )
        || ( ! ( reason & CB_SC_P_IO ) )
        || ( ! ( reason & CB_SC_P_CD ) )
      )
      reg_cmd_info.failbits |= FAILBIT8;

   /*
   ** Done...
   ** Read the output registers and trace the command.
   ** Also put the command packet in the trace.
   */

   trace_command();
   trc_cht_pkt();

   /*
   ** For interrupt mode, restore the INT 7x vector.
   */

   int_restore_int_vect();

   /*
   ** mark end of PI cmd in low level trace
   */

   trc_llt( 0, 0, TRC_LLT_E_PI );

   /*
   ** All done.  The return values of this function are described in
   ** ATAIO.H.
   */

   reg_atapi_max_bytes = 32768L;    /* reset max bytes */
   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/* end ataioreg.c */
